/****
  Minimax.hpp
  
  */
  
#ifndef _MINIMAX_HPP_
#define _MINIMAX_HPP_

#include <time.h>
#include <assert.h>
#include "board.hpp"
#include "endgamecx.h"

#define MAX_INT 2147483647
#define MIN_INT -2147483648
#define LARGE_FLOAT 1.0e35
#define SMALL_FLOAT -1.0e35
#define HUGE_FLOAT 2.0e38
#define TINY_FLOAT -2.0e38

#define HARD_WIRED_FIRST_MOVE 37  // F-5 position
#define HARD_WIRED_FIRST_MOVE_BOARD_FLIPPED 34  // C-5

/* Manually given weight -- ideally should be automatically learnt */
#define DISC_WEIGHT_TEST -0.26  // was used when testing (no longer used)
#define DISC_WEIGHT_STAGE_1 -0.22
#define DISC_WEIGHT_STAGE_2 -0.10
#define DISC_WEIGHT_STAGE_3 -0.05
#define DISC_WEIGHT_STAGE_4  0.001
#define STAGE_1_END    24
#define STAGE_2_START  30
#define STAGE_2_END    36
#define STAGE_3_START  42
#define STAGE_3_END    48
#define STAGE_4_START  50

#define ESTIMATED_CORNER_WORTH 15.0  // wild estimate
#define NEAR_TO_CORNER_NEG_DIAG 0.5
#define NEAR_TO_CORNER_NEG_EDGE 0.1
#define NEAR_TO_CORNER_NEG_DIAG_COEFF 0.01
#define NEAR_TO_CORNER_NEG_EDGE_COEFF 0.002
#define DENOMINATOR_EXTRA_DOF 0.99  // avoiding divide by 0
#define RANDOM_FACTOR_CONSTANT 0.05

// need to be experimented for best values
#define USE_MOVE_ORDERING 1
#define DO_USE_QUIESCENCE_SEARCH  0 // to be implemented!
#define MIN_DEPTH_FOR_ORDERING 6  // min depth to use ordering
#define MAX_ORDERING_SEARCH_DEPTH 6 // max depth of an ordering search
#define ORDERING_SEARCH_DEPTH_OFFSET 4 // how many plies fewer is ordering search than normal search
#define MIN_EMPTIES_FOR_ORDERING 12 // min number of empty squares left for doing search ordering

// test/debug 
#define DEBUG_MINIMAX 0
#define AB_DEBUG 0
#define DEEP_DEBUG 0
#define CURRENT_DEBUG 0
#define CURRENT_DEBUG2 0

// produce some statistics
#define COUNT_PRUNING 1

extern char searchDepth;
extern char originalSearchDepth;
extern char bruteForceDepth; // for approaching the end of game.
extern char mpcDepth;
extern bool winLarge;
extern bool loseSmall;
extern char randomnessLevel;
extern bool useAndersson;  // use Andersson's sophisticated end game solver

extern bool boardFlipped;
extern bool showDots;  // simply for output


char getMinimaxMove(Board *board, bool *lm);
float getMin(char *a, char lastMove, char color, char depth, char passes, 
             char prevmaxDOF, char prevminDOF, //Mobility in previous getMax and getMin;
             char selfPieces, char oppPieces, float alpha, float beta);
float getMax(char *a, char lastMove, char color, char depth, char passes, 
             char prevmaxDOF, char prevminDOF, 
             char selfPieces, char oppPieces, float alpha, float beta);

void copyBoardArray(char *to, char *from);
bool canFlip(char *a, char place, char dir, char color, char oppcolor);
bool legalMove(char *a, char color, char place);
char findLegalMoves(char *a, char color, char *legalMoves);
char tryMove(char *a, char color, char place, char *flipped);
inline void undo(char *a, char color, char place, char *flipped, char nFlipped);

int evaluateEndGame(char selfPieces, char oppPieces);
int evaluateEndGameWDL(char selfPieces, char oppPieces);
float evaluateBoard(char *a, char forWhom, char whoseTurn, char prevmaxDOF, 
                    char prevminDOF, char selfPieces, char oppPieces);

void printBoardArray(char *a);
int strongEndGameSolve(char *a, char color, char selfPieces, char oppPieces, 
                      char prevNotPass, float alpha, float beta);

#endif
